// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_Resource_hpp
#define anna_core_mt_Resource_hpp

#include <anna/core/mt/Mutex.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/util/Millisecond.hpp>

namespace anna {

namespace xml {
class Node;
}

/**
   Clase que modela un recurso de acceso compartido.
*/
class Resource : public Mutex {
public:
  /**
     Constructor.
     \param name Nombre logico de este recurso remoto.
     \param mode Modo de actuacion de esta instancia en modo ST cuando montamos una seccion critca sobre este objeto.
     En modo MT siempre sera Mode::Normal, es decir abre una seccion critica sobre este objeto, que bloqueara a
     cualquier otro thread que intente acceder a el.
  */
  Resource(const std::string& name) :
    Mutex(),
    a_name(name),
    a_isEnabled(true),
    a_timeStamp(0)
  {;}

  /**
     Devuelve el nombre logico de este recurso remoto.
     @return El nombre logico de este recurso remoto.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Devuelve el estado referente a la habilitacion de este recurso.

     La habilitacion o inhabilitacion responde a causas externas a este recurso, es decir, puede ser
     que el recurso estaria preparado pero por algun motivo deseamos marcarlo como "no usable".

     Para que un recurso pueda ser usado tanto este metodo como #isAvailable deberan retornar \em true.

     \return \em true Si el recurso esta habilitado o \em false en otro caso.
  */
  bool isEnabled() const throw() { return a_isEnabled; }

  /**
  * Obtiene el valor de la marca de tiempo asociada a este recurso. Se usa desde las clases
  * de reparto de carga, para marca el momento en que ha sido desactivado el recurso, lo permite
  * volver a comprobar el recurso cada cierto periodo de tiempo.
  *
  * \param timeStamp Marca de tiempo asociada al recurso.
  */
  void setTimeStamp(const Millisecond &timeStamp) throw() { a_timeStamp = timeStamp; }

  /*
  * Devuelve la marca de tiempo asociada al recurso.
  * \return la marca de tiempo asociada al recurso.
  */
  const Millisecond &getTimeStamp() const throw() { return a_timeStamp; }

  /**
     Operador de comparacion.
     \param name Nombre con el que comparar.
     \return Devuelve \em true si el nombre recibido como parametro coincide con el
     indicado en el constructor de esta instancia o \em false en caso contrario.
  */
  bool operator == (const std::string& name) const throw() { return a_name == name; }

  /**
     Operador de comparacion.
     \param other Instancia con la que comparar.
     \return Devuelve \em true si el nombre recibido como parametro coincide con el
     indicado en el constructor de esta instancia o \em false en caso contrario.
  */
  bool operator == (const Resource& other) const throw() { return a_name == other.a_name; }

  /**
     Habilita el uso de este recurso.
     Si no se indica lo contrario un recurso remoto siempre se encuentra habilitado, pero no tiene por que estar
     disponible.
     \warning La invocacion a este metodo deberia hacerse en modo exclusivo.
  */
  void enable() throw(RuntimeException)  { Guard guard(this, "Resource"); a_isEnabled = true;   }

  /**
     Evita el uso de este recurso.
     \warning La invocacion a este metodo deberia hacerse en modo exclusivo.
  */
  void disable() throw(RuntimeException)  { Guard guard(this, "Resource"); a_isEnabled = false;  }

  /**
     Devuelve una cadena con la informacin referente a este recurso remoto.
     @return Una cadena con la informacin referente a este recurso remoto.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve el estado de disponibilidad de este recurso.

     El recurso no esta preparado para ser usado debido a causas internas.

     Para que un recurso pueda ser usado tanto este metodo como #isEnabled deberan retornar \em true.

     @return \em false si el recurso no esta preparado para ser usado debido a causas internas o
      \em true en otro caso.
  */
  virtual bool isAvailable() const throw(RuntimeException) = 0;

private:
  const std::string a_name;
  bool a_isEnabled;
  Millisecond a_timeStamp;
};

}

#endif
