// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_ThreadManager_hpp
#define anna_core_mt_ThreadManager_hpp

#include <string>

#include <anna/core/mt/SafeRecycler.hpp>
#include <anna/core/mt/Thread.hpp>

#include <anna/core/core.hpp>

namespace anna {

class Semaphore;

/**
   Gestor de threads. Optimiza la creacin y uso de los threads. Como ventaja adicional limita
   el numero de threads que puede ser creados.
*/
class ThreadManager : SafeRecycler <Thread> {
public:
  /**
     Modos de funcionamiento
     \see ThreadManager
  */
  struct Mode {
    enum _v {
      None,
      Unlimit, /**< Permite crear tantos threads como sea necesario. */
      ExceptionWhenFull, /**< Si no hay ningun thread libre lanza una excepcion */
      LockWhenFull /**< Si no hay ningun thread libre el metodo que invoca a #createThread queda
         bloqueado hasta que algun otro thread termine su tarea. */
    };

    anna_declare_enum(Mode);
  };

  typedef iterator thread_iterator;

  /**
     Constructor.
     @param name Nombre logico del la gestor de threads.
     \param mode Modo de funcionamiento de ester gestor de threads.
     @param maxSize Numero maximo de threads que seran creados.
     \param flags Flags aplicados a los threads creados por este gestor.
  */
  ThreadManager(const char* name, const Mode::_v mode, const int maxSize, const int flags = Thread::Flag::None);

  /**
     Constructor.
     Crea un gestor de thread que puede crecer con un nmero ilimitado de threads.
     @param name Nombre logico del la gestor de threads.
     \param flags Flags aplicados a los threads creados por este gestor.
  */
  ThreadManager(const char* name, const int flags = Thread::Flag::None);

  /**
     Destructor.
  */
  virtual ~ThreadManager();

  /**
   * Devuelve el nombre asociado a este gestor.
   * \return el nombre asociado a este gestor.
   */
  const std::string& getName() const throw() { return a_name; }

  /**
     Obtiene la instancia de un thread. El thread slo se liberar de forma automtica cuando se termine la ejecucin del mismo.
     \return La instancia de un thread.

     \warning Solo debe haber un nico punto de creacin de thread's por cada instancia de esta clase.
  */
  Thread* createThread() throw(RuntimeException);

  /**
   * Bloquea el proceso hasta que todos los threads lanzados por este gestor hayan terminado su ejecucin.
   * \warning \li No debera invocarse desde un thread que haya sido creado por este gestor.
   *          \li Los threads deberan crearse con el flag Thread::Flag::Joinable.
   */
  void join() throw(RuntimeException);

  /**
     Devuelve un iterador al comienzo de la lista de threads.
     \return un iterador al comienzo de la lista de threads.
  */
  thread_iterator thread_begin() throw() { return begin(); }

  /**
     Devuelve un iterador al final de la lista de threads.
     \return un iterador al final de la lista de threads.
  */
  thread_iterator thread_end() throw() { return end(); }

  /**
     Devuelve el thread referenciado por el iterador recibido como parametro.
     \param ii Iterador.
     \return el thread referenciado por el iterador recibido como parametro.
  */
  static Thread* thread(thread_iterator ii) throw() { return SafeRecycler <Thread>::data(ii); }

  /**
   * Devuelve una cadena con la informacin relevante sobre este gestor de threads
   * \return una cadena con la informacin relevante sobre este gestor de threads
   */
  std::string asString() const throw();

private:
  const std::string a_name;
  const Mode::_v a_mode;
  const int a_maxSize;
  Semaphore* a_semaphore;
  const int a_threadFlags;
  bool a_destroying;

  // Se invoca cuando el thread detecta que va a terminar y que tiene una factoria asociada.
  void releaseThread(Thread* thread) throw(RuntimeException);

  friend class Thread;
};

} //namespace anna

#endif

