// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_oam_Handler_hpp
#define anna_core_oam_Handler_hpp


#include <anna/core/RuntimeException.hpp>

// STL
#include <string>

#include <anna/core/oam/defines.hpp>

#include <cstdarg>


namespace anna {

namespace oam {

class Module;


/**
* Class used to manage OAM events and registration over a OAM module
*/
class Handler {

protected:

  /**
  * Event for counters generated at the module provided. Base implementation trace the event based on #Module::getDefaultCounterDescription.
  *
  * Any re-implementation should invoke base class method.
  *
  * @param module Reference OAM module.
  * @param type Counter enum-identification within the own context/module.
  * @param amount Units increased.
  *
  * @return Boolean counter data for 'type' when registered, NULL if not found
  */
  virtual const counter_data_t *counterEvent(const Module *module, const int & type, const int & amount) const throw();

  /**
  * Event for alarms generated at the module provided. Base implementation trace the event based on #Module::getDefaultAlarmDescription,
  * replacing tags '__s__' and '__d__' by '<\%s>' and '<\%d>' (for texts and numbers) if they exists, and parsing with
  * dynamic variables passed on activateAlarm()/cancelAlarm() prototypes.
  *
  * Any re-implementation should invoke base class method.
  *
  * @param module Reference OAM module.
  * @param textPreffix Dynamic alarm modification regarding text preffix.
  * @param textSuffix Dynamic alarm modification regarding text suffix.
  * @param textSeparator Dynamic alarm modification regarding text sections separator.
  * @param Activation activation/cancellation indicator.
  * @param type Alarm enum-identification within the own context/module.
  * @param argList Optional parsing data for dynamic-composed text.
  *
  * @return Boolean alarm data for 'type' when registered, NULL if not found
  */
  virtual const alarm_data_t *alarmEvent(const Module *module, const char *textPreffix, const char *textSuffix, char textSeparator, bool activation, const int & type, va_list argList) const throw();


public:

  /**
   * Constructor
   */
  Handler() {;}

  /**
   * Destructor
   */
  ~Handler() {;}


  /**
  * Invokes module counter registration procedure.
  * Default implementation does nothing specific
  *
  * @param module Reference OAM module
  * @see #anna::oam::Module::registerCounter
  */
  virtual void registerCounter(Module *module, const int &type, const std::string &description, const int &offset) throw(anna::RuntimeException) {;}

  /**
  * Invokes module alarm registration procedure.
  * Default implementation does nothing specific
  *
  * @param module Reference OAM module
  * @see #anna::oam::Module::registerAlarm
  */
  virtual void registerAlarm(Module *module, const int &type, const std::string &description, const int &externalId, const std::string &dynamicVariablesCSL, const int &activationId, const int &cancellationId = -1) throw(anna::RuntimeException) {;}


  friend class Module;
};

}
}

#endif

