// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_tracing_TraceMethod_hpp
#define anna_core_tracing_TraceMethod_hpp

#include <anna/core/tracing/Logger.hpp>

namespace anna {

/**
  Trazas de funciones.

  Graba una traza en el historico de operaciones por cada entrada y salida de un determinado
  metodo.
*/
class TraceMethod {
public:
  /**
    Constructor.
    @param className Nombre de la clase desde la que se invoca.
    @param functionName Nombre de la funcion desde la que se invoca a este constructor.
    @param fromFile Nombre del fichero desde el que se invoca a este constructor. Normalmente
    sera el indicado por la macro de compilacion __FILE__.
    @param fromLine Numero de linea del fichero desde la que se invoca a este constructor.
    Normalmente sera el indicado por la macro de compilacion __LINE__.

    \warning Se pueden sustituir los dos ultimos valores por la macro ANNA_FILE_LOCATION.
  */
  TraceMethod(const char* className, const char* functionName, const char* fromFile, const int fromLine) :
    a_level(Logger::Debug),
    a_className(className),
    a_functionName(functionName),
    a_fromFile(fromFile),
    a_ok(false) {
    if((a_ok = Logger::isActive(a_level)) == true) {
      std::string text(className);
      text += "::";
      text += functionName;
      Logger::write(a_level, text, "begin", fromFile, fromLine);
    }
  }

  /**
    Constructor.
    @param level Nivel usado en la traza del mtodo
    @param className Nombre de la clase desde la que se invoca.
    @param functionName Nombre de la funcion desde la que se invoca a este constructor.
    @param fromFile Nombre del fichero desde el que se invoca a este constructor. Normalmente
    sera el indicado por la macro de compilacion __FILE__.
    @param fromLine Numero de linea del fichero desde la que se invoca a este constructor.
    Normalmente sera el indicado por la macro de compilacion __LINE__.

    \warning Se pueden sustituir los dos ultimos valores por la macro ANNA_FILE_LOCATION.
  */
  TraceMethod(const Logger::Level level, const char* className, const char* functionName, const char* fromFile, const int fromLine) :
    a_level(level),
    a_className(className),
    a_functionName(functionName),
    a_fromFile(fromFile) {
    if((a_ok = Logger::isActive(level)) == true) {
      std::string text(className);
      text += "::";
      text += functionName;
      Logger::write(level, text, "begin", fromFile, fromLine);
    }
  }

  /**
    Constructor.
    @param methodName Nombre del mtodo desde la que se invoca, debera incluir el nombre de la clase.
    @param fromFile Nombre del fichero desde el que se invoca a este constructor. Normalmente
    sera el indicado por la macro de compilacion __FILE__.
    @param fromLine Numero de linea del fichero desde la que se invoca a este constructor.
    Normalmente sera el indicado por la macro de compilacion __LINE__.

    \warning Se pueden sustituir los dos ultimos valores por la macro ANNA_FILE_LOCATION.
  */
  TraceMethod(const char* methodName, const char* fromFile, const int fromLine) :
    a_level(Logger::Debug),
    a_className(methodName),
    a_functionName(NULL),
    a_fromFile(fromFile),
    a_ok(false) {
    if((a_ok = Logger::isActive(a_level)) == true)
      Logger::write(a_level, methodName, "begin", fromFile, fromLine);
  }

  /**
    Constructor.
    @param level Nivel usado en la traza del mtodo.
    @param methodName Nombre del mtodo desde la que se invoca, debera incluir el nombre de la clase.
    @param fromFile Nombre del fichero desde el que se invoca a este constructor. Normalmente
    sera el indicado por la macro de compilacion __FILE__.
    @param fromLine Numero de linea del fichero desde la que se invoca a este constructor.
    Normalmente sera el indicado por la macro de compilacion __LINE__.

    \warning Se pueden sustituir los dos ultimos valores por la macro ANNA_FILE_LOCATION.
  */
  TraceMethod(const Logger::Level level, const char* methodName, const char* fromFile, const int fromLine) :
    a_level(level),
    a_className(methodName),
    a_functionName(NULL),
    a_fromFile(fromFile),
    a_ok(false) {
    if((a_ok = Logger::isActive(level)) == true)
      Logger::write(level, methodName, "begin", fromFile, fromLine);
  }

  /**
     Destructor.
  */
  ~TraceMethod() {
    if(a_ok == true && Logger::isActive(a_level) == true) {
      std::string text(a_className);

      if(a_functionName != NULL) {
        text += "::";
        text += a_functionName;
      }

      Logger::write(a_level, text, "end", a_fromFile, 0);
    }
  }

private:
  const Logger::Level a_level;
  const char* a_className;
  const char* a_functionName;
  const char* a_fromFile;
  bool a_ok;
};

} //namespace anna

#endif

