// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_CommandLine_hpp
#define anna_core_util_CommandLine_hpp

#include <stdlib.h>

#include <vector>

#include <anna/core/functions.hpp>
#include <anna/core/Singleton.hpp>

namespace anna {

class RuntimeException;

namespace xml {
class Node;
}


/**
   Command line parser helper for our application. It's close to GNU-style, supporting
   single letter (single hyphen) and long argument names (double hyphen). No bare hyphen
   or double-dash end of parsing separator are supported. No positional arguments are supported.
*/
class CommandLine : public Singleton <CommandLine> {

  /* returns first no-leading hyphen position; -1 is error */
  static int removeLeadingHyphens(std::string &argv) throw();

public:
  /**
     Define los tipos de argumento

  */
  struct Argument { enum Type { Mandatory = 0, Optional};  };

  // Accesores
  /**
     @return la lista de cadenas indicadas en la linea de comandos al ejecutar este programa.
     Mientras que no invoquemos al metodo #initialize devolvera NULL.
  */
  const char** getArgv() const throw() { return a_argv; }

  /**
     @return El numero de parametros indicados en la linea de comandos al ejecutar este programa.
  */
  int getArgc() const throw() { return a_argc; }

  //
  // Metodos
  //
  /**
     Establece la informacion necesaria para analizar la linea de comandos
     indicada por el usuario. Debe invocarse antes que cualquier otro metodo
     relacionado con la obtencion/comprobacion de valor de la linea de comandos.

     Recibe una copia de los parametros que los recibidos por el metodo 'main'.

     @param argv Conjunto de cadenas que se reciben de la linea de comandos.
     @param argc Numero de cadenas recibidas.
     @param positionalArguments Enables positional arguments. An offset will be applied to start command-line interpretation.
     These positional arguments are mandatory, and the user could retrieve their values through #getPositional. By default no
     positional arguments are specified.
  */
  void initialize(const char** argv, const int argc, int positionalArguments = 0) throw(RuntimeException);

  /**
     Register an argument name in our application

     @param argumentExpression Argument name, or comma-separated set with both short and long argument names. For example 'v,version', 'h,help', or simply 'f' or 'file'. If both,
     provided, one of them shall be a single letter and the other will be a word. In other case, nothing will be registered. Command line arguments stands for -<single letter option>
     and --<word option> when proceed. If NULL provided, nothing is done.
     @param type Argument type. See Variable#Type.
     @param comment Argument explanation.
     @param needValue If our argument has an additional associated value, this will be true. False in other case (flags).
   */
  void add(const char* argumentExpression, Argument::Type type, const char* comment, const bool needValue = true) throw();

  /**
     Gets a positional argument. There must be registered or NULL will be returned.

     @param position Argument position from 1 to N

     @return Value of mandatory positional argument with position provided
  */
  const char *getPositional(int position) const throw() {
    const char *result = NULL;
    if ((position > 0) && (position <= a_positionalArguments)) result = a_argv[position];
    return result;
  }

  /**
     Obtiene el valor asociado al argumento recibido como parametro.
     El valor devuelto puede ser NULL en caso de que el argumento no sea
     obligatorio y no este en la linea de comandos.
     Si el argumento es obligatorio y no este en la linea de comandos o
     no tiene valor asociado la ejecucion del programa TERMINA inmediatamente.

     @param argumentExpression You should look for the registered expression (#add), internally tokenized if needed.
     @param exitOnFault Indica el funcionamiento del metodo en caso de que el
     argumento solicitado no halla sido indicado. Si el parametro no existe
     si vale @em true la aplicacion terminara, si vale @em false devolvera NULL.

     @return Valor asociadoal argumento recibido como parametro. Puede ser NULL.
  */
  const char* getValue(const char* argumentExpression, const bool exitOnFault = true) throw();

  /**
     Obtiene el valor asociado al argumento recibido, pero convirtiendo a
     numero el valor obtenido por #getValue.

     @param argumentExpression You should look for the registered expression (#add), internally tokenized if needed.

     @return Valor numerico del valor devuelto por #getValue.
  */
  int getIntegerValue(const char* argumentExpression) throw() { return atoi(getValue(argumentExpression)); }

  /**
     Comprueba si el argumento recibido como parametro esta presente en la linea de
     comandos.

     @param argumentExpression You should look for the registered expression (#add), internally tokenized if needed.

     @return true si el argumento esta en la linea de comandos y false en otro caso.
   */
  bool exists(const char* argumentExpression) throw() { return (getValue(argumentExpression, false) != NULL) ? true : false; }

  /**
    Comprueba la linea de comandos del programa para verificar que coincide con los argumentos
    registrados por nuestra aplicacion:

    @li Verifica que los parametros obligatorios estan en la linea de comandos.
    @li Verifica que los valores de los argumento son correctos de forma que si un parametro
    debe llevar un valor asociado este esta presente y que si no debe llevarlo no esta.
    El orden en que aparezcan los argumento en la linea de comandos es indiferente a la hora de
    hacer las comprobacion.

    Si hay algun fallo en la linea de comandos establecida al ejecutar el programa visualiza un
    resumen con los parametros soportados y la ejecucion del programa finaliza.
  */
  void verify() throw(RuntimeException);


  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  std::string asString() const throw();

  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  xml::Node* asXML(xml::Node* parent) const throw();


private:
  Mutex a_mutex;

  class Variable {
  public:
    // Constructors
    Variable(const std::string &name1, const std::string &name2, const Argument::Type type, const char* comment, const bool needValue = true) :
      a_name1(name1), a_name2(name2), a_type(type), a_comment(comment), a_needValue(needValue), a_isOn(false), a_value(NULL) {
    }
    virtual ~Variable() { if(a_value) free(a_value); }

    // Accesores
    const std::string& getName1() const throw() { return a_name1; }
    const std::string& getName2() const throw() { return a_name2; }
    std::string getHelpExpression() const throw();
    const char* getValue() const throw() { return a_value; }
    const char* getComment() const throw() { return a_comment; }
    bool getNeedValue() const throw() { return a_needValue; }
    bool getIsOn() const throw() { return a_isOn; }
    Argument::Type getType() const throw() { return a_type; }

    // Modificadores
    void setValue(const char* value) throw() { a_value = (value == NULL) ? NULL : strdup(value); }
    void setIsOn(const bool isOn) throw() { a_isOn = isOn; }

    // Metodos
    std::string asString() const throw();

  protected:
    std::string a_name1, a_name2;
    Argument::Type a_type;
    const char* a_comment;
    bool a_needValue;
    bool a_isOn;
    char* a_value;
  };

  const char **a_argv;
  int a_argc;
  bool a_wasParsed;
  std::vector <Variable*> a_arguments;
  int a_positionalArguments;

  CommandLine() : a_argv(NULL), a_argc(0), a_positionalArguments(0)  {;}

  bool analize() throw();
  const Variable* search(const char *argumentExpression) const throw();
  void printUsage() const throw();

  friend class Singleton <CommandLine>;
};

}

#endif

