// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_Message_hpp
#define anna_http_Message_hpp

#include <anna/core/util/Recycler.hpp>

#include <anna/comm/Message.hpp>

#include <anna/http/Header.hpp>
#include <anna/http/Version.hpp>

namespace anna {

namespace http {

namespace parser {
class Abstract;
}

/**
   Clase base de los mensajes HTTP segun la RFC 2616.
*/
class Message : public comm::Message {
public:
  typedef Recycler <Header>::iterator header_iterator;
  typedef Recycler <Header>::const_iterator const_header_iterator;

  /**
   * Para poder acceder a todos los mtodos de anna::comm::Message;;setBody cuando
   * se invocan desde un anna::http::Message (o heredados).
   */
  using comm::Message::setBody;

  /**
     Tipos de mensaje HTTP.
     \see Message
  */
  struct Type {
    enum _v {
      Request, /**< Peticion HTTP */
      Response /**< Respuesta HTTP */
    };
  };

  /**
     Devuelve el tipo de este mensaje HTTP.
     \return el tipo de este mensaje HTTP.
  */
  Type::_v getType() const throw() { return a_type; }

  /**
     Devuelve la version del protocolo HTTP usada en el mensaje.
     \return la version del protocolo HTTP usada en el mensaje.
  */
  const std::string& getVersion() const throw() { return a_version; }

  /*
   * Devuelve el enumerado que identifica la versin HTTP usada en el mensaje.
   * \return el enumerado que identifica la versin HTTP usada en el mensaje.
   */
  Version::_v getVersionAsEnum() const throw() { return Version::asEnum(a_version); }

  /**
   * Devuelve los parmetros extras que pueden haber estado contenidos en los bloques
   * de datos, en caso de que el mensaje se haya recibido como un "Transfer-Encoding: chunked".
   * \return Los parmetros extras que pueden haber estado contenidos en los bloques
   */
  const std::string& getExtraParameters() const throw() { return a_extraParameters;  }

  /**
     Establece la version a codificar en el mensaje.
     \param version Texto con la version HTTP a usar, HTTP/1.1, por ejemplo.
  */
  void setVersion(const std::string& version) throw() { a_version = version; }

  /**
   * Establece la versin a codificar en el mensaje.
   * \param version Identificador de la versin.
   */
  void setVersion(const Version::_v version) throw() { a_version = Version::asCString(version); }

  /**
   * Establece el cuerpo de este mensaje con el contenido del documento XML correspondiente al
   * nodo XML recibido como parmetro.
   *
   * Tambin establece la cabecera \em Content-Type.
   *
   * \param node Nodo XML que contiene el documento XML.
   */
  comm::Message* setBody(const xml::Node* node) throw(RuntimeException);

  /**
     Crea una nueva cabecera en este mensaje.
     \param type Tipo de cabecera a crear.
     \return La instancia de la nueva cabecera.
  */
  Header* createHeader(const Header::Type::_v type)
  throw(RuntimeException) {
    return a_headers.create()->initialize(type);
  }

  /**
     Crea una nueva cabecera en este mensaje.
     \param name Nombre de la cabecera a crear.
     \return La instancia de la nueva cabecera.
  */
  Header* createHeader(const std::string& name)
  throw(RuntimeException) {
    return a_headers.create()->initialize(name);
  }

  /**
     Devuelve la instancia de la primera cabecera que coincide con el tipo
     recibido como parametro. Puede ser NULL si el mensaje no contiene ninguna
     cabecera del tipo solicitado.
     \param type Tipo de cabecera buscado.
     \return La primera cabecera que coincide con el tipo recibido como parametro.
  */
  const Header* find(const Header::Type::_v type) const throw() {
    return const_cast <Message*>(this)->find(type);
  }

  /**
     Devuelve la instancia de la primera cabecera que coincide con el tipo
     recibido como parametro. Puede ser NULL si el mensaje no contiene ninguna
     cabecera del tipo solicitado.
     \param type Tipo de cabecera buscado.
     \return La primera cabecera que coincide con el tipo recibido como parametro.
  */
  Header* find(const Header::Type::_v type) throw();

  /**
     Devuelve la instancia de la primera cabecera de tipo Header::Category::Extension cuyo
     nombre coincide con el recibido como parametro. Puede ser NULL.
     \param name Nombre de la extension buscada.
     \return La primera cabecera que coincide con el tipo recibido como parametro.
  */
  Header* find(const char* name) throw();

  /**
     Inicializa el contenido de este mensaje. Libera las cabeceras y el cuerpo del mensaje.
  */
  void clear() throw() { a_headers.clear(); clearBody(); a_extraParameters.clear();}

  /**
     Devuelve un iterador al comienzo de la lista de cabeceras.
     \return un iterador al comienzo de la lista de cabeceras.
  */
  header_iterator header_begin() throw() { return a_headers.begin(); }

  /**
     Devuelve un iterador al final de la lista de cabeceras.
     \return un iterador al final de la lista de cabeceras.
  */
  header_iterator header_end() throw() { return a_headers.end(); }

  /**
     Devuelve un iterador al comienzo de la lista de cabeceras.
     \return un iterador al comienzo de la lista de cabeceras.
  */
  const_header_iterator header_begin() const throw() { return a_headers.begin(); }

  /**
     Devuelve un iterador al final de la lista de cabeceras.
     \return un iterador al final de la lista de cabeceras.
  */
  const_header_iterator header_end() const throw() { return a_headers.end(); }

  /**
     Codifica este mensaje.
     \return El bloque de datos que contiene el mensaje codificado.
  */
  const DataBlock& code() throw(anna::RuntimeException);

  /**
     Devuelve una cadena con toda la informacion relevante de este objeto.
     \return una cadena con toda la informacion relevante de este objeto.
  */
  virtual std::string asString() const throw() = 0;

  /**
     Devuelve la instancia de la cabecera a la que referencia el iterator
     recibido como parametro.
     \return la instancia de la cabecera a la que referencia el iterator
     recibido como parametro.
  */
  static Header* header(header_iterator& ii) throw() { return Recycler<Header>::data(ii); }

  /**
     Devuelve la instancia de la cabecera a la que referencia el iterator
     recibido como parametro.
     \return la instancia de la cabecera a la que referencia el iterator
     recibido como parametro.
  */
  static const Header* header(const_header_iterator& ii) throw() { return Recycler<Header>::data(ii); }

protected:
  /**
     Constructor.
     \param type Tipo de mensaje.
  */
  Message(const Type::_v type) : comm::Message(StatusCodeBuffer::Reserve),
    a_type(type), a_version("HTTP/1.1")
  {;}

private:
  const Type::_v a_type;
  std::string a_version;
  Recycler <Header> a_headers;
  std::string a_extraParameters;

  void codeLine(const std::string& line) throw(RuntimeException);

  void appendExtraParameter(const std::string& extraParameter) throw() {
    a_extraParameters += ' ';
    a_extraParameters += extraParameter;
  }

  virtual std::string codeStartLine() const throw(anna::RuntimeException) = 0;

  friend class parser::Abstract;
};


}
}

#endif
