// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/RuntimeException.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/app/functions.hpp>

#include <anna/comm/CongestionController.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/comm/Communicator.hpp>

using namespace std;
using namespace anna;

// static
const int comm::CongestionController::MaxPendingBytes = 64 * 1024;
const Millisecond comm::CongestionController::DelayTrace(5000);

// Si pasa este tiempo sin recibir peticiones de consejo => entiende que el proceso ha dejado de recibir carga => inicializar las estadsticas
//static
const Millisecond comm::CongestionController::HeartBeat(1000);

comm::CongestionController::CongestionController() :
  a_avgWorkload("comm::CongestionController::AvgWorkload"),
  a_timeTrace(0),
  a_maxPendingBytes(UnusedPendingBytes),
  a_incomingSocketCounter(0),
  a_tickTime(0) {
  srand(time(NULL));
  a_percentage [0] = 25;
  a_percentage [1] = 50;
  a_percentage [2] = 85;
  a_percentage [3] = 99;
  a_limit = 0;
  setLimit(DefaultLimit);
  setMode(Mode::Auto);
}

void comm::CongestionController::setLimit(const int limit)
throw() {
  Guard guard(a_mutex, "comm::CongestionController (setLimit)");

  if(limit < 0  || limit > 100) {
    LOGWARNING(
      Logger::warning(
        functions::asString("comm::CongestionController::setLimit | Limit %d out of range [0,100]", limit),
        ANNA_FILE_LOCATION
      );
    );
    return;
  }

  if(a_limit == limit)
    return;

  int congestionArea = 100 - (a_limit = limit);
  a_discardLevel [0] = (congestionArea * a_percentage [0] / 100) + a_limit;
  a_discardLevel [1] = (congestionArea * a_percentage [1] / 100) + a_limit;
  a_discardLevel [2] = (congestionArea * a_percentage [2] / 100) + a_limit;
  a_discardLevel [3] = 100;
  a_messageCounter = a_discardedCounter = 0;
  LOGINFORMATION(
    string msg = functions::asString(
                   "comm::CongestionController::setLimit | Limit: %d | ", limit
                 );

    for(int i = 0; i < MaxLevel; i ++)
    msg += functions::asString("(Level %d: %d%%) ", i + 1, a_discardLevel [i]);
    Logger::information(msg, ANNA_FILE_LOCATION);
  );
}

void comm::CongestionController::setMaxPendingBytes(const int maxPendingBytes)
throw(RuntimeException) {
  if(maxPendingBytes == UnusedPendingBytes) {
    a_maxPendingBytes = UnusedPendingBytes;
    return;
  }

//   const int minimum (Communicator::getReceivingChunkSize ());
//   const int maximum (min (minimum << 1, CongestionController::MaxPendingBytes));
  const int minimum(0);
  const int maximum(CongestionController::MaxPendingBytes);

  if(maxPendingBytes < minimum || maxPendingBytes > maximum) {
    string msg("comm::CongestionController::setMaxPendingBytes | ");
    msg += functions::asString("MaxPendingBytes (%d bytes) must be between %d bytes and %d bytes", maxPendingBytes, minimum, maximum);
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_maxPendingBytes = maxPendingBytes;
}

/*
 * Se invoca desde [Tzzz] seguramente que con el clientSocket protegido por la seccin crtica establecida
 * por su manejador asociado.
 */
comm::CongestionController::Advice::_v comm::CongestionController::getAdvice(const ClientSocket& clientSocket)
throw() {
  Guard guard(a_mutex, "comm::CongestionController::getAdvice");

  if(a_limit == 0) {
    a_messageCounter ++;
    return Advice::Process;
  }

  Advice::_v result = Advice::Process;
  int workload = calculeWorkload(clientSocket);
  int discard = 0;

  // Si le da la vuelta al marcador
  if(++ a_messageCounter == 0) {
    a_messageCounter = 1;
    a_discardedCounter = 0;
    a_avgWorkload.clear();
  }

  const Millisecond now = functions::millisecond();

  if((now - a_tickTime) > HeartBeat)
    a_avgWorkload.clear();

  a_tickTime = now;

  if(workload != -1) {
    a_avgWorkload += workload;

    if(a_effectiveMode == Mode::Global)
      workload = a_avgWorkload;

    LOGDEBUG(
      std::string msg("comm::CongestionController::getAdvice | ");
      msg += a_avgWorkload.asString();
      msg += functions::asText(" | inmediate workload: ", workload);
      Logger::debug(msg, ANNA_FILE_LOCATION)
    );

    if(workload  < a_limit) {
      result = Advice::Process;
      discard = 0;
    } else {
      result = Advice::Discard;
      discard = 100;

      for(int i = 0; i < MaxLevel; i ++) {
        if(workload > a_discardLevel [i])
          continue;

        result = ((rand() % 101) > (discard = a_percentage [i])) ? Advice::Process : Advice::Discard;
        break;
      }
    }
  }

  if(result == Advice::Process) {
    LOGDEBUG(
      const int ratio = (a_messageCounter - a_discardedCounter) * 100 / a_messageCounter;
      string msg = functions::asString(
                     "Mode: %s | Limit: %d%% | Workload: %d%% | Filter: %d%% | Ratio: %d%% | Result: Process",
                     ((a_effectiveMode == Mode::Local) ? "Local" : "Global"), a_limit, workload, discard, ratio
                   );
      Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  } else {
    a_discardedCounter ++;
    const int ratio = (a_messageCounter - a_discardedCounter) * 100 / a_messageCounter;

    if(Logger::isActive(Logger::Debug) == true) {
      string msg = functions::asString(
                     "Mode: %s | Limit: %d%% | Workload: %d%% | Filter: %d%% | Ratio (%u/%u): %d%% | Result: Discard",
                     ((a_effectiveMode == Mode::Local) ? "Local" : "Global"), a_limit, workload, discard, a_messageCounter, a_discardedCounter, ratio
                   );
      Logger::debug(msg, ANNA_FILE_LOCATION);
    } else if(Logger::isActive(Logger::Warning)) {
      Millisecond now = functions::millisecond();

      if((now - a_timeTrace) >= DelayTrace) {
        string msg = functions::asString(
                       "Mode: %s | Limit: %d%% | Workload: %d%% | Filter: %d%% | Ratio (%u/%u): %d%% | Result: Discard",
                       ((a_effectiveMode == Mode::Local) ? "Local" : "Global"), a_limit, workload, discard, a_messageCounter, a_discardedCounter, ratio
                     );
        Logger::warning(msg, ANNA_FILE_LOCATION);
        a_timeTrace = now;
      }
    }
  }

  return result;
}

void comm::CongestionController::incrementIncomingSocket()
throw(RuntimeException) {
  Guard guard(a_mutex, "comm::CongestionController::incrementIncomingSocket");
  a_incomingSocketCounter ++;
}

void comm::CongestionController::decrementIncomingSocket()
throw(RuntimeException) {
  if(a_incomingSocketCounter ==  0)
    return;

  Guard guard(a_mutex, "comm::CongestionController::incrementIncomingSocket");
  a_timeTrace = 0;

  if(a_incomingSocketCounter == 0) {
    a_avgWorkload.clear();
    return;
  }

  const int average = a_avgWorkload;

  unsigned int individualWorkload = average / a_incomingSocketCounter;

  if(-- a_incomingSocketCounter < 0)
    a_incomingSocketCounter = 0;

  if(a_incomingSocketCounter > 0)
    a_avgWorkload.setValue(individualWorkload * a_incomingSocketCounter * a_incomingSocketCounter, a_incomingSocketCounter);
  else
    a_avgWorkload.clear();
}

int comm::CongestionController::calculeWorkload(const ClientSocket& clientSocket) const
throw() {
  int maxSize;

  if(a_maxPendingBytes == UnusedPendingBytes)
    maxSize = clientSocket.getReceiveBufferSize();
  else
    maxSize = a_maxPendingBytes;

  // El tamao actual tiene en cuenta lo que ha cargado en memoria, ms lo que hay en el I/O pendiente de cargar
  const int size = clientSocket.getTotalPendingBytes();
  const int result = (maxSize == 0) ? -1 : (size * 100) / maxSize;
  LOGDEBUG(
    string msg = functions::asString("fd: %d | Size (%d%%): %d/%d", clientSocket.getfd(), result, size, maxSize);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  return result;
}

xml::Node* comm::CongestionController::asXML(xml::Node* parent) const
throw() {
  static const char* modetxt [] = { "Auto", "Local", "Global" };
  xml::Node* result = parent->createChild("comm.CongestionController");
  result->createAttribute("Limit", a_limit);
  result->createAttribute("Mode", modetxt [a_mode]);
  result->createAttribute("EffectiveMode", modetxt [a_effectiveMode]);

  if(a_maxPendingBytes != UnusedPendingBytes)
    result->createAttribute("MaxPendingBytes", a_maxPendingBytes);

  xml::Node* w;
  Workload current = getAccumulatedWorkload();
  w = result->createChild("CurrentStatus");
  w->createAttribute("Level", getLevel(current));
  w->createAttribute("Workload", getLoad(current));
  w->createAttribute("Process", a_messageCounter);
  w->createAttribute("Discard", a_discardedCounter);
  w->createAttribute("IncominSockets", a_incomingSocketCounter);
  w = result->createChild("Levels");
  w->createAttribute(
    "Ratio", ((a_messageCounter == 0) ? 0 : (a_messageCounter - a_discardedCounter) * 100 / a_messageCounter)
  );
  xml::Node* node;
  node = w->createChild("Level");
  node->createAttribute("Id", 0);
  node->createAttribute("Limit", a_limit - 1);
  node->createAttribute("Discarding", 0);

  for(int i = 0; i < MaxLevel; i ++) {
    node = w->createChild("Level");
    node->createAttribute("Id", i + 1);
    node->createAttribute("Limit", a_discardLevel [i]);
    node->createAttribute("Discarding", a_percentage [i]);
  }

  return result;
}

comm::CongestionController::Workload comm::CongestionController::getAccumulatedWorkload() const
throw() {
  Workload result;
  const Millisecond now = functions::millisecond();

  if((now - a_tickTime) > HeartBeat) {
    CongestionController* _this = const_cast <CongestionController*>(this);
    _this->a_avgWorkload.clear();
    _this->a_tickTime = now;
  }

  result.second = a_avgWorkload;

  if(result.second < a_limit)
    result.first = 0;
  else {
    result.first = 4;

    for(int ii = 0; ii < MaxLevel; ii ++) {
      if(result.second > a_discardLevel [ii])
        continue;

      result.first = ii + 1;
      break;
    }
  }

  return result;
}


comm::CongestionController::Workload comm::CongestionController::getCurrentWorkload(const comm::ClientSocket& clientSocket) const
throw() {
  Workload result;
  result.second = calculeWorkload(clientSocket);

  if(result.second < a_limit)
    result.first = 0;
  else {
    result.first = 4;

    for(int ii = 0; ii < MaxLevel; ii ++) {
      if(result.second > a_discardLevel [ii])
        continue;

      result.first = ii + 1;
      break;
    }
  }

  return result;
}
